try:
    import pegtree as pg
    from pegtree.visitor import ParseTreeVisitor
except ModuleNotFoundError:
    import os
    os.system('pip3 install pegtree')
    import pegtree as pg
    from pegtree.visitor import ParseTreeVisitor


def fix(tree):
    a = [tree.epos_]
    for t in tree:
        a.append(fix(t).epos_)
    for key in tree.keys():
        a.append(fix(tree.get(key)).epos_)
    #print(repr(tree), a)
    tree.epos_ = max(a)
    return tree


def fix2(s):
    if not isinstance(s, str):
        s = str(fix(s))
    open = s.count('(')
    close = s.count(')')
    if open > 0 and open > close:
        return s+(')' * (open-close))
    open = s.count('[')
    close = s.count(']')
    if open > 0 and open > close:
        return s+(']' * (open-close))
    return s


parser = pg.generate(pg.grammar('inspect.pegtree'))


class Flatten(ParseTreeVisitor):
    def __init__(self):
        ParseTreeVisitor.__init__(self)

    def parse(self, s: str, bufs=None):
        tree = parser(s)
        self.bufs = [] if bufs is None else bufs
        self.visit(tree)
        return self.bufs

    def acceptName(self, tree):
        self.bufs.append(dict(
            type='var',
            name=str(tree),
        ))

    def acceptInfix(self, tree):
        self.bufs.append(dict(
            type='infix',
            name=str(tree.name).strip(),
            left=str(fix2(tree.left)),
            right=str(fix2(tree.right)),
        ))
        self.visit(tree.left)
        self.visit(tree.right)

    def acceptApplyExpr(self, tree):
        self.bufs.append(dict(
            name=fix2(tree.name),
            type='app',
        ))
        self.visit(tree.params)

    def acceptGetExpr(self, tree):
        if tree.recv == 'Name':
            self.bufs.append(dict(
                name=str(tree.recv),
                property=str(tree.name),
                type='module',
            ))
        self.bufs.append(dict(
            name=str(tree.name),
            callee=fix2(tree.recv),
            type='prop',
        ))

    def acceptMethodExpr(self, tree):
        if tree.recv == 'Name':
            self.bufs.append(dict(
                name=str(tree.recv),
                prop=str(tree.name),
                type='module',
            ))
        self.bufs.append(dict(
            name=str(tree.name),
            callee=fix2(tree.recv),
            type='app',
        ))
        self.visit(tree.params)

    def acceptIndexExpr(self, tree):
        self.bufs.append(dict(
            sub=fix2(tree.index),
            callee=fix2(tree.recv),
            type='sub',
        ))
        self.visit(tree.recv)
        self.visit(tree.index)

    def acceptUndefined(self, tree):
        for t in tree:
            self.visit(t)
        for key in tree.keys():
            self.visit(tree.get(key))


f = Flatten()


def filter(results, type='sub', name=None):
    rs = []
    if name is None:
        for r in results:
            if r['type'] == type:
                rs.append(r)
    else:
        for r in results:
            if r['type'] == type and r['name'] == name:
                rs.append(r)
    return rs


def filter2(results, name, defined='callee'):
    rs = []
    for r in results:
        if defined in r and r['name'] == name:
            rs.append(r)
    return rs


def find_infix(lines, infix='+'):
    if lines is not None:
        for line in lines:
            results = filter(f.parse(line), type='infix', name=infix)
            if len(results) > 0:
                return results
    return []


def find_sub(lines):
    if lines is not None:
        for line in lines:
            results = filter(f.parse(line), type='sub')
            if len(results) > 0:
                return results
    return []


def find_app(lines):
    if lines is not None:
        for line in lines:
            results = filter(f.parse(line), type='app')
            if len(results) > 0:
                return results
    return []


def find_callee(lines, name):
    if lines is not None:
        for line in lines:
            results = filter2(f.parse(line), name=name)
            if len(results) > 0:
                return results
    return []


def find_name(lines, name):
    if lines is not None:
        for line in lines:
            results = filter2(f.parse(line), name=name, defined='name')
            if len(results) > 0:
                return results
    return []


def inspect_name(code, lines, slots):
    if lines is None:
        return ''
    return ''


def inspect_callee(code, lines, slots):
    if lines is None:
        return ''

    results = find_callee(lines, slots['name'])
    # print(slots)
    # print(results, 'callee' in results)
    for result in results:
        if 'callee' in result:
            slots['callee'] = result['callee']
            if result.get('type', 'prop') == 'prop':
                slots['pyname'] = 'プロパティ'
            else:
                slots['pyname'] = '関数' if 'module' in slots['error'] else 'メソッド'
            # print(slots)
            return '_ext'
    return ''


def inspect_infix(code, lines, slots):
    if lines is None:
        return ''
    results = find_infix(lines, slots['name'])
    # print(slots)
    # print(results)
    for result in results:
        if 'left' in result:
            slots['left'] = result['left']
            slots['right'] = result['right']
            # print(slots)
            return '_ext'
    return ''


def inspect_funcname(code, lines, slots):
    if lines is None:
        return ''
    results = find_app(lines)
    # print(slots)
    # print(results)
    for result in results:
        if 'name' in result:
            slots['name'] = result['name']
            # print(slots)
            return '_ext'
    return ''


def inspect_index(code, lines, slots):
    if lines is None:
        return ''
    results = find_sub(lines)
    # print(slots)
    # print(results)
    for result in results:
        if 'sub' in result:
            slots['index'] = result['sub']
            slots['callee'] = result['callee']
            # print(slots)
            return '_ext'
    return ''


if __name__ == '__main__':
    print(find_sub(['a(df["A"])+a[0]']))
    print(find_app(['a(2)+b()']))
    print(find_infix(['a()+a[0]-1'], infix='-'))
    print(find_callee(['math.sin(math.pi)'], name='sin'))
    print(find_callee(['math.sin(math.pi)'], name='pi'))
    print(find_name(['pd.read_csv()'], name='pd'))
    print(find_name(['int()'], name='int'))




try:
    from .webui import kogi_login, kogi_help, kogi_translate, kogi_print
except ModuleNotFoundError:
    pass

from .runner import kogi, corgi
#from .atcoder import input, print, _check_atcoder, _run_judge
#kogi_add_option(_check_atcoder, _run_judge)





PYTHON_SIMPLENAME = {
    'NoneType': 'None',
    'bool': '論理値',
    'int': '整数',
    'float': '浮動小数点数',
    'complex': '複素数',
    'str': '文字列',
    'bytes': 'バイト列',
    'list': 'リスト',
    'tuple': '組',
    'dict': '辞書',
    'set': 'セット',
    'ndarray': '配列',
    'DataFrame': 'データフレーム',
    'module': 'モジュール',
    'builtin_function_or_method': 'ビルトイン関数',
    'function': '関数',
    'type': '型もしくはクラス',
}


# HTML render

def render(s: str, type='type', render_html=True) -> str:
    if type.startswith('type'):
        s = PYTHON_SIMPLENAME.get(s, f'{s}型')
    if not render_html:
        return s
    if type.startswith('type'):
        return f'<span class="type">{s}</span>'
    return f'<b style="color: #c3c; font-family: monospace;">{s}</b>'


def render_astype(value: object, render_html=True) -> str:
    return render(type(value).__name__, 'type', render_html=render_html)


def render_value(value: object, render_html=True) -> str:
    if not render_html:
        return str(value)
    if hasattr(value, '_repr_html_'):
        return value._repr_html_()
    if hasattr(value, '__len__') and len(value) > 8:
        return render(str(value[:8]), 'value', render_html=render_html) + '...(以下、省略)'
    return render(repr(value), 'value', render_html=render_html)



_PYTYPE = {
    'NoneType': 'None',
    'bool': '論理値',
    'int': '整数',
    'float': '実数/浮動小数点数',
    'complex': '複素数',
    'str': '文字列',
    'bytes': 'バイト列',
    'list': 'リスト',
    'tuple': '組/タプル',
    'dict': '辞書/マッピング',
    'set': '集合/セット',
    'ndarray': '配列/NumPy配列',
    'DataFrame': '表データ/データフレーム',
    'module': 'モジュール',
    'builtin_function_or_method': '関数/ビルトイン関数',
    'function': '関数',
    'type': '型もしくはクラス',
}


def _span(s, css_class, return_html=True):
    if return_html:
        return f'<span class="{css_class}">{s}</span>'
    return s


def _kogi_typename(value, return_html=True):
    typename = type(value).__name__
    if typename in _PYTYPE:
        return _span(_PYTYPE[typename], "type", return_html)
    return _span(typename, "type", return_html) + 'クラス'


def _kogi_typename(type_name, suffix=''):
    if type_name in _PYTYPE:
        return _simple(_PYTYPE[type_name]) + suffix
    else:
        return type_name + _simple('型/クラス')


def _kogi_inspect(v):
    type_name = type(v).__name__
    d = {'typeid': type_name, 'type': _kogi_typename(type_name)}
    if hasattr(v, '__next__'):
        d['iterable'] = hasattr(v, '__next__')
    if hasattr(v, '__len__'):
        d['len'] = len(v)
    if hasattr(v, '__name__'):
        d['name'] = v.__name__
    return d


def _html_symbol(s, return_html=True):
    if return_html:
        return f'<span class="symbol" style="color: #c3c; font-family: monospace;">{s}</span>'
    return s


def _html_value(s, return_html=True):
    if len(s) > 100:
        s = s[:100] + '...(以下、省略)'
    if return_html:
        return f'<span class="value" style="font-family: monospace">{s}</span>'
    return s


def _kogi_say_names(names, msgs, return_html=True):
    for d in names:
        if 'value' in d:
            symbol = _html_symbol(d['id'], return_html=True)
            typename = _html_type(d['type'], return_html=True)
            value = d['value']
            if hasattr(value, '_repr_html_'):
                value = value._repr_html_()
            else:
                value = _html_value(repr(value))
            msg = f'{symbol}は、{typename}。値は{value}'
            msgs.append(msg)




from IPython.core.magic import register_cell_magic
from .runner import kogi_run, kogi_add_option
from .atcoder_impl import print, input, _check_atcoder, _run_judge

kogi_add_option(_check_atcoder, _run_judge)


@register_cell_magic
def atcoder(option, code):
    kogi_run(code, option, run_cell=_run_judge)



import difflib
import requests
from bs4 import BeautifulSoup
import builtins
from IPython.display import display, HTML
from .logger import kogi_print, log, send_log

_lines = None
_outputs = None


def input(s=''):
    global _lines
    if _lines is not None and len(_lines) > 0:
        return _lines.pop(0)
    else:
        _lines = None
    return builtins.input(s)


def print(*a, **kw):
    if _outputs is not None:
        sep = kw.get('sep', ' ')
        end = kw.get('end', '\n')
        s = sep.join([str(s) for s in a]) + end
        _outputs.append(s)
    else:
        builtins.print(*a, **kw)


def _get_problemid(problem):
    problem, _, _ = problem.partition('&')
    _, _, problem = problem.rpartition('/')
    problem = problem.lower()
    num = 'a'
    if problem[-1] in 'abcdefgh':
        num = problem[-1]
        problem = problem[:-1].replace('_', '').replace('-', '')
    return f'{problem}_{num}'


def _get_url(problem):
    if '/' in problem:
        return problem
    problem = problem.lower()
    num = 'a'
    if problem[-1] in 'abcdefgh':
        num = problem[-1]
        problem = problem[:-1].replace('_', '').replace('-', '')
    return f'https://atcoder.jp/contests/{problem}/tasks/{problem}_{num}'


SAMPLE = {}


def _get_sample(problem):
    if '/' in problem:
        problem = problem[problem.rfind('/')+1:]
    problem = problem.lower()
    if '_' in problem:
        problem, num = problem.split('_')
    else:
        num = problem[-1].lower()
        problem = problem[:-1].lower()
    pid = f'{problem}_{num}'
    if pid in SAMPLE:
        return SAMPLE[pid]
    response_text = requests.get(
        url=f"https://atcoder.jp/contests/{problem}/tasks/{problem}_{num}").text
    html = BeautifulSoup(response_text, "lxml")
    d = {}
    for a in html.find_all("section"):
        # print(a)
        if a.h3 and a.pre:
            key = a.h3.text.replace('\r\n', '\n')
            value = a.pre.text.replace('\r\n', '\n')
            d[key] = value
    SAMPLE[pid] = d
    return d


def _check_atcoder(option):
    try:
        if 'atcoder' in option:
            d = _get_sample(option)
            return True
    except Exception as e:
        print('問題が読み込めません', e)
    return False


_COLOR_HTML_DIC = {
    'yellow': '<span style="background: pink">',
    'red': '<span style="background: lightblue">',
    'end': '</span>'
}


def _display_diff(ground_truth, target):
    _CDIC = _COLOR_HTML_DIC

    if ground_truth == target:
        return

    d = difflib.Differ()
    diffs = d.compare(ground_truth, target)

    result = ''
    for diff in diffs:
        status, _, character = list(diff)
        if status == '-':
            character = _CDIC['red'] + character + _CDIC['end']
        elif status == '+':
            character = _CDIC['yellow'] + character + _CDIC['end']
        else:
            pass
        result += character

    display(
        HTML(f'<h4>差分</h4><div style="white-space: pre-wrap;">{result}</div>'))


JUDGE_CSS = '''
<style>
.parent {
  background-color: #edebeb;
  width: 100%;
  //height: 150px;
}
textarea {
  width: 100%; 
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  //height:120px; 
  font-size: large;
  outline: none;           /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
.box18{  // ひだり
  //padding: 0.2em 0.5em;
  //margin: 2em 0;
  color: #565656;
  background: #ffeaea;
  background-size: 150%;
  background-repeat: no-repeat;
  background-position: top right;
  background-color:rgba(255,255,255,0.8);
  //box-shadow: 0px 0px 0px 10px #ffeaea;
  border: dashed 2px #ffc3c3;
  //border-radius: 8px;
}
.box16{
    background: repeating-linear-gradient(-45deg, #D5FFB0, #D5FFB0 3px,#ffffff 3px, #ffffff 7px);
}
.box17{
    background-image: repeating-linear-gradient(45deg, rgba(255, 0, 0, .3), rgba(255, 0, 0, .3) 10px, #FFEFF7 10px, #FFEFF7 20px);
}
.box24 {
    position: relative;
    padding: 0.5em 0.7em;
    margin: 2em 0; background: #6f4b3e;
    color: white; font-weight: bold;
}
.box23:after {
    position: absolute;
    content: '';
    top: 100%; left: 30px;
    border: 15px solid transparent;
    border-top: 15px solid #6f4b3e;
    width: 0; height: 0;
}
</style>
'''

JUDGE_HTML = '''
<div class="parent" style="clear:both;">
<h4>{title}</h4>
<pre>{input}</pre>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="input">実行結果</label>
<textarea class="box18" style="height:{height}" readonly>{output}</textarea>
</div>
<div style="float: left; width: 48%">
<label class="box24" for="outout">正解例</label>
<textarea class="box18" style="height:{height}" readonly>{sample}</textarea>
</div>
</div>
<p style="clear:both;"></p>
'''

AC_HTML = '''
<div style="clear:both;">
AtCoderでACを取るためには、<b>制約条件</b>を満たす全ての入力にパスするようにプログラムする必要があります。<br/>
もう一度、確認してから<a href="{url}" target="atcoder">提出</a>しましょう。
</div>
'''


def _run_judge(code, problem):
    global _lines, _outputs
    d = _get_sample(problem)
    if len(d) == 0:
        kogi_print('問題データが読み込めません')
        res = get_ipython().run_cell(code)
        res.raise_error()
        return
    try:
        ac = 0
        display(HTML(JUDGE_CSS))
        for key in ['入力例 1', '入力例 2', '入力例 3']:
            if key not in d:
                continue
            data = {'title': key, 'input': d[key]}
            # display(HTML(sample_html))
            _lines = [s for s in d[key].split('\n') if len(s) > 0]
            _outputs = []
            res = get_ipython().run_cell(code)
            res.raise_error()
            key = key.replace('入力', '出力')
            data['sample'] = d[key]
            data['output'] = ''.join(_outputs)
            data['box'] = 'box16' if data['sample'] == data['output'] else 'box17'
            ac += 1 if data['sample'] == data['output'] else 0
            lines = max(data['output'].count('\n'),
                        data['sample'].count('\n'))+1
            data['height'] = '240px' if lines > 10 else f'{lines*24}px'
            display(HTML(JUDGE_HTML.format(**data)))
        display(HTML(AC_HTML.format(url=_get_url(problem))))
        log(type='atcoder', problem=_get_problemid(problem), ac=ac, code=code)
    finally:
        _lines = None
        _outputs = None



def print_nop(*args):
    pass


def listfy(s: str) -> tuple:
    """リストにして返す
    Returns:
        s: タプルにして返す
    """
    if not isinstance(s, (list, tuple)):
        return (s,)
    return s


def _ZEN():
    _ZA = ''.join(chr(ord('Ａ')+c) for c in range(26))
    _Za = ''.join(chr(ord('ａ')+c) for c in range(26))
    _Z0 = ''.join(chr(ord('０')+c) for c in range(10))
    _Z = '\u3000「」！＂＃＄％＆＇（）＊＋，－．／：；＜＝＞＠［＼］＾＿｀｛｜｝'
    return _Z + _ZA + _Za + _Z0


def _HAN():
    _HA = ''.join(chr(ord('A')+c) for c in range(26))
    _Ha = ''.join(chr(ord('a')+c) for c in range(26))
    _H0 = ''.join(chr(ord('0')+c) for c in range(10))
    _H = '\u0020\'\'!"#$%&\'()*+, -./:<=>@[\\]^_`{|}'
    return _H + _HA + _Ha + _H0


_ZEN2HAN = str.maketrans(_ZEN(), _HAN())


def zen2han(s: str) -> str:
    return s.translate(_ZEN2HAN)



from setuptools import setup, find_packages

'''
python3 -m unittest
vim setup.py
rm -rf dist/
python3 setup.py sdist bdist_wheel
twine upload --repository pypi dist/*
'''


def _requires_from_file(filename):
    return open(filename).read().splitlines()


setup(name="kogi",
      version="0.0.2",
      license='MIT',
      author='Kimio Kuramitsu',
      description="Kogi Programming AI",
      url="https://github.com/kkuramitsu/kogi",
      packages=['kogi'],
      #package_dir={"": "src"},
      package_data={'kogi': ['./*.pegtree', 'data/*.*']},
      install_requires=_requires_from_file('requirements.txt'),
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Framework :: IPython',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Intended Audience :: Education',
      ],
      )



import pegtree as pg
import pandas as pd
from .logger import kogi_print, print_nop
from .utils import zen2han


TYPE_PREFIX = {
    'type': 'クラス',
    'bool': 'ブール値',
    'int': '',
    'float': '',
    'dict': '辞書',
    'list': 'リスト',
    'tuple': 'タプル',
    'function': '関数',
    'set': 'セット',
    'deque': 'デック',
    'Counter': 'カウンタ',
    'ndarray': '配列',
    'DataFrame': 'データフレーム',
    'Series': 'データ列',
    'Symbol': 'シンボル',
    'AssocOp': '式',
}

FILEEXT = {
    'txt': 'テキストファイル',
    'csv': 'CSVファイル',
    'tsv': 'TSVファイル',
}


COLUMN_NAMES = set()


def update_columns(columns):
    COLUMN_NAMES.update(columns)


def detect_string_prefix(name, s):
    if s in COLUMN_NAMES:
        return 'カラム'
    _, sep, ext = s.rpartition('.')
    if sep == '.' and ext in FILEEXT:
        return FILEEXT[ext]
    if '{' in s and '}' in s:
        return '書式'
    return '文字列'


def detect_prefix(name, v):
    if isinstance(v, str):
        return detect_string_prefix(name, v)
    if isinstance(v, pd.DataFrame):
        update_columns(list(v.columns))
    typename = type(v).__name__
    if typename in TYPE_PREFIX:
        return TYPE_PREFIX[typename]
    supername = (type(v).__base__).__name__
    if supername in TYPE_PREFIX:
        return TYPE_PREFIX[supername]
    return ''  # '{typename}'

# code, prefix, key, index = prefix_index_code(str(fix(t)), index)


def prefix_code(code, index, always_policy=False):
    try:
        v = eval(code)
    except:
        if always_policy:
            return code, '', f'<e{index}>', index+1
        return code, '', '', index
    return code, detect_prefix(code, v), f'<e{index}>', index+1


PEG = '''
Start = {
    (Code _ / Chunk)*
    #Statement
}

Chunk = {
    . (!Code .)* #Chunk
}

Code = '`' { (!'`' .)+ #Code } '`' / { '<' (!'>' .)+ '>' #Special } / Expression

Expression =
  Suffix {^ OP Suffix #Binary }*

OP =
  / "==" / "!=" / "<=" / ">=" / "<>"
  / "**" / "//" / ">>" / "<<<" / ">>"
  / "+" / "-" / "*" / "/" / "%" / "="

Suffix = Primary _Postfix*

_Postfix =
  / {^ "(" Expression? ("," Expression )* ")" #App}
  / {^ "[" Expression? (([:,] _) Expression )* "]" #Index}

Primary =
  / Group
  / String
  / Name
  / Number

Group = 
  / "(" Expression ")"
  / { "[" Expression? ("," Expression)* ","? "]"  #List }
  / { "(" Expression? ("," Expression)* ","? ")"  #Tuple }

Name = 
  { [A-Za-z_] [A-Za-z_.0-9]* #Name } _

Number = 
  / { [0-9]+ ('.' [0-9]+)? #Number } _
  / { '.' [0-9]+ #Number } _

String =
  / { 'f'? '"'  (!'"' .)* '"' #String } _
  / { 'f'? "'" (!"'" .)*  "'" #String } _
'''

peg = pg.grammar(PEG)
parser = pg.generate(peg)


def _fix(tree):
    a = [tree.epos_]
    for t in tree:
        a.append(_fix(t).epos_)
    for key in tree.keys():
        a.append(_fix(tree.get(key)).epos_)
    tree.epos_ = max(a)
    return tree


def _replace_expression(s: str, always_policy=False):
    tree = parser(s)
    ss = []
    vars = {}
    index = 0
    # print(repr(tree))
    for t in tree:
        tag = t.getTag()
        if tag == 'Chunk' or tag == 'Special':
            ss.append(str(t))
        else:
            code = str(_fix(t))
            code, prefix, key, index = prefix_code(
                code, index, always_policy=always_policy)
            if key == '':
                ss.append(str(t))
            else:
                vars[key] = code
                ss.append(f'{prefix}{key}')
    return ''.join(ss), vars


def _replace_special(s: str, vars: dict) -> str:
    if isinstance(s, (list, tuple)):
        return tuple(_replace_special(x, vars) for x in s)
    for key in vars:
        s = s.replace(key, vars[key])
    return s


def _translate(s: str) -> str:
    return f'trans: {s}'


def compose(translate=_translate):
    def replace_around(s, always_policy=False, print=print_nop):
        s = zen2han(s)
        if s.startswith('trans'):
            s, vars = _replace_expression(s, always_policy=always_policy)
            print('[BEFORE]', s)
            s = translate(s)
            print('[AFTER]', s)
            return _replace_special(s, vars)
        else:
            return translate(s)
    return replace_around


if __name__ == '__main__':
    tr = compose()
    tr('「a」をみる', always_policy=True, print=print)
